/* Copyright(C) 2004 Brazil

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.
  
  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.
  
  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
#include <stdio.h>
#include <string.h>
#include "senna.h"
#include "lib/senna_in.h"
#include "inv.h"

/* copied from inv.c */

#define NEXT_ADDR(p) (((byte *)(p)) + sizeof *(p))

/* encode/decode */

#define ENCODE(v,p) \
{ \
  uint8_t *_p = (uint8_t *)p; \
  uint32_t _v = v; \
  if (_v < 0x8f) { \
    *_p++ = _v; \
  } else if (_v < 0x408f) { \
    _v -= 0x8f; \
    *_p++ = 0xc0 + (_v >> 8); \
    *_p++ = _v & 0xff; \
  } else if (_v < 0x20408f) { \
    _v -= 0x408f; \
    *_p++ = 0xa0 + (_v >> 16); \
    *_p++ = (_v >> 8) & 0xff; \
    *_p++ = _v & 0xff; \
  } else if (_v < 0x1020408f) { \
    _v -= 0x20408f; \
    *_p++ = 0x90 + (_v >> 24); \
    *_p++ = (_v >> 16) & 0xff; \
    *_p++ = (_v >> 8) & 0xff; \
    *_p++ = _v & 0xff; \
  } else { \
    *_p++ = 0x8f; \
    *((uint32_t *)_p) = _v; \
    _p += sizeof(uint32_t); \
  } \
  p = _p; \
}

#define DECODE(v,p) \
{ \
  uint8_t *_p = (uint8_t *)p; \
  uint32_t _v = *_p++; \
  switch (_v >> 4) { \
  case 0x08 : \
    if (_v == 0x8f) { \
      _v = *((uint32_t *)_p); \
      _p += sizeof(uint32_t); \
    } \
    break; \
  case 0x09 : \
    _v = (_v - 0x90) * 0x100 + *_p++; \
    _v = _v * 0x100 + *_p++; \
    _v = _v * 0x100 + *_p++ + 0x20408f; \
    break; \
  case 0x0a : \
  case 0x0b : \
    _v = (_v - 0xa0) * 0x100 + *_p++; \
    _v = _v * 0x100 + *_p++ + 0x408f; \
    break; \
  case 0x0c : \
  case 0x0d : \
  case 0x0e : \
  case 0x0f : \
    _v = (_v - 0xc0) * 0x100 + *_p++ + 0x8f; \
    break; \
  } \
  v = _v; \
  p = _p; \
}

#define SKIP(p) \
{ \
  uint8_t *_p = (uint8_t *)p; \
  uint32_t _v = *_p++; \
  switch (_v >> 4) { \
  case 0x08 : \
    if (_v == 0x8f) { \
      _p += sizeof(uint32_t); \
    } \
    break; \
  case 0x09 : \
    _p += 3; \
    break; \
  case 0x0a : \
  case 0x0b : \
    _p += 2; \
    break; \
  case 0x0c : \
  case 0x0d : \
  case 0x0e : \
  case 0x0f : \
    _p += 1; \
    break; \
  } \
  p = _p; \
}

typedef struct {
  uint16_t step;
  uint16_t jump;
} buffer_rec;

#define BUFFER_REC_AT(b,pos) ((buffer_rec *)(b) + (pos))

/* copied from inv.c */

struct timeval tv0;

inline static void
t0(void)
{
  gettimeofday(&tv0, NULL);
}

inline static void
ts(const char *msg)
{
  struct timeval tv;
  long long int t1;
  gettimeofday(&tv, NULL);
  t1 = (tv.tv_sec - tv0.tv_sec) * CLOCKS_PER_SEC + (tv.tv_usec - tv0.tv_usec);
  printf("%08lld %s\n", t1, msg);
}

int
main(int argc, char **argv)
{
  sen_id tid;
  sen_index *i;
  sen_inv_cursor *c;
  unsigned char key[0x1000];
  if (argc < 2) {
    fputs("usage: mtest indexname\n", stderr);
    return -1;
  }
  t0();
  sen_init();
  ts("init");
  if (!(i = sen_index_open(argv[1]))) {
    fprintf(stderr, "index open failed (%s)\n", argv[1]);
    return -1;
  }
  ts("open");
  while (!feof(stdin) && fgets(key, 0x1000, stdin)) {
    int bdf = 0, cdf = 0, size = 0, posts = 0;
    key[strlen(key) - 1] = '\0';
    t0();
    tid = sen_sym_at(i->lexicon, key);
    ts("sym_at");
    printf("tid=%d\n", tid);
    c = sen_inv_cursor_open(i->inv, tid);
    ts("cursor open");
    if (!c) { continue; }
    while (c->nextb) {
      uint32_t tf, lrid = c->pb.rid, lsid = c->pb.sid; /* for check */
      buffer_rec *br = BUFFER_REC_AT(c->buf, c->nextb);
      c->bp = NEXT_ADDR(br);
      DECODE(c->pb.rid, c->bp);
      DECODE(c->pb.sid, c->bp);
      if (lrid > c->pb.rid || (lrid == c->pb.rid && lsid >= c->pb.sid)) {
	sen_log("brokend!! (%d:%d) -> (%d:%d)", lrid, lsid, c->pb.rid, c->pb.sid);
      }
      c->nextb = br->step;
      DECODE(tf, c->bp);
      if (tf & 1) { DECODE(c->pb.score, c->bp); } else { c->pb.score = 0; }
      c->pb.rest = c->pb.tf = tf >> 1;
      c->pb.pos = 0;
      bdf++;
    }
    printf("buffer df=%d\n", bdf);
    ts("buffer scan");
    size = c->cpe - c->cp;
    while (c->cp < c->cpe) {
      uint32_t tf, gap;
      DECODE(gap, c->cp);
      c->pc.rid += gap;
      if (gap) { c->pc.sid = 0; }
      DECODE(gap, c->cp);
      c->pc.sid += gap;
      DECODE(tf, c->cp);
      if (tf & 1) { DECODE(c->pc.score, c->cp); } else { c->pc.score = 0; }
      c->pc.rest = c->pc.tf = tf >> 1;
      c->pc.pos = 0;
      // while (c->cp < c->cpe && c->pc.rest--) { SKIP(c->cp); }
      posts += c->pc.tf;
      while (c->cp < c->cpe && c->pc.rest--) { DECODE(gap, c->cp); }
      cdf++;
    }
    printf("chunk size=%d df=%d posts=%d\n", size, cdf, posts);
    ts("chunk scan");
    sen_inv_cursor_close(c);
    ts("cursor close");
  }
  return 0;
}
